"""
    This file implements the necessary classes for data fetcher class
    DataFetcher forks several processes that request data and transfer them
    through memory mapped files
"""
from threading import Thread
from multiprocessing import Process
from caraml.zmq.communicator import ZmqClient, ZmqServer
from caraml.utils.serializer import get_serializer, get_deserializer
from caraml.inmemory import inmem_serialize, inmem_deserialize

_CARAML_TERMINATE_FETCHER = '_CARAML_TERMINATE_FETCHER'


def _get_new_task_message():
    return {
        'type': 'new-task',
    }


def _get_response_message(request, data):
    return {
        'type': 'response',
        'request': request,
        'data': data,
    }


class DataFetcherWorker(Process):
    def __init__(self, master_port, remote_host=None, remote_port=None,
                 handler=None, serializer=None, deserializer=None):
        """
        Override this to setup states in the main process
        """
        Process.__init__(self)
        self.master_port = master_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.handler = handler
        self.serializer = get_serializer(serializer)
        self.deserializer = get_deserializer(deserializer)

    def run(self):
        # You must initilize the transmission channel AFTER you fork off
        self.remote = ZmqClient(host=self.remote_host, port=self.remote_port, 
                                serializer=self.serializer,
                                deserializer=self.deserializer)
        self.master = ZmqClient(host='127.0.0.1',
                                port=self.master_port,
                                serializer='pickle',
                                deserializer='pickle')
        while True:
            task = self.master.request(_get_new_task_message())
            if task == _CARAML_TERMINATE_FETCHER:
                return
            else:
                response = self.remote.request(task)
                if self.handler is not None:
                    response = self.handler(response)
                name = inmem_serialize(response)
                self.master.request(_get_response_message(task, name))


class DataFetcher(Thread):
    def __init__(self, handler,
                 remote_host,
                 remote_port,
                 requests,
                 worker_comm_port,
                 remote_serializer='pyarrow',
                 remote_deserialzer='pyarrow',
                 n_workers=2,
                 worker_handler=None,
                 ):
        """
            TODO: automatically find a port
            TODO: drain and exit

            Forks @n_worker processes, send @requests to
            remote_host:remote_port and process responses
            with handler

        Args:
            handler: function(request, reponse) what to do with response in the
                main process
            remote_host, remote_port: host/port for workers to connect to
            requests: A generator of requests to send to remote
            worker_comm_port: local port for communication with workers
            remote_serializer: how to serialize requests to remote
            remote_deserializer: how to deserialize response from remote
            n_workers: number of processes to fetch data
            worker_handler: how to process data on worker processes
                (i.e. compute intensive tasks that still return
                pyarrow serializable data)
        """
        Thread.__init__(self)
        self.handler = handler
        self.n_workers = n_workers
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.remote_serializer = remote_serializer
        self.remote_deserialzer = remote_deserialzer
        self.requests_iter = iter(requests)

        self.worker_comm_port = worker_comm_port
        self.n_workers = n_workers
        self.worker_handler = worker_handler

    def run(self):
        # Use receiver here to rate-limit the workers, using pull-push causes
        # a lot of messages to be held in system socket cache
        # https://stackoverflow.com/questions/22613737/how-could-i-set-hwm-in-the-push-pull-pattern-of-zmq
        self.server = ZmqServer(host='127.0.0.1',
                                port=self.worker_comm_port,
                                serializer='pickle',
                                deserializer='pickle')
        self.workers = []

        self.active_workers = 0
        for i in range(self.n_workers):
            worker = DataFetcherWorker(master_port=self.worker_comm_port,
                                       remote_host=self.remote_host,
                                       remote_port=self.remote_port,
                                       handler=self.worker_handler,
                                       serializer=self.remote_serializer,
                                       deserializer=self.remote_deserialzer,
                                       )
            worker.start()
            self.active_workers += 1
            self.workers.append(worker)

        while True:
            message = self.server.recv()
            if message['type'] == 'new-task':
                try:
                    self.server.send(self.requests_iter.__next__())
                except StopIteration:
                    self.server.send(_CARAML_TERMINATE_FETCHER)
                    self.active_workers -= 1
                    if self.active_workers == 0:
                        break
            elif message['type'] == 'response':
                self.server.send('ack')
                request = message['request']
                data = inmem_deserialize(message['data'])
                self.handler(request, data)

        for worker in self.workers:
            worker.join()



from .communicator import (
    ZmqTimeoutError,
    ZmqSocket,
    ZmqPusher,
    ZmqPuller,
    ZmqSender,
    ZmqReceiver,
    ZmqClient,
    ZmqServer,
    ZmqPub,
    ZmqSub)
from .inmem_data_fetcher import DataFetcher
from .proxy import ZmqProxy, ZmqProxyThread, ZmqProxyProcess



"""
Provides abstractions over zmq sockets
"""
from threading import Thread
import zmq
import nanolog as nl
from caraml.utils.serializer import get_serializer, get_deserializer, str2bytes


logger = nl.Logger.create_logger(
    __name__,
    stream='stdout',
    time_format='MD HMS',
    show_level=True,
)


class ZmqTimeoutError(Exception):
    """
    Raised when a socket operation times out
    """

    def __init__(self):
        super().__init__('Request Timed Out')


class ZmqSocket(object):
    """
        Wrapper around the data that we use to initialize zmq sockets.
        Use establish() to actually create the socket
    """
    SOCKET_TYPES = {
        'PULL': zmq.PULL,
        'PUSH': zmq.PUSH,
        'PUB': zmq.PUB,
        'SUB': zmq.SUB,
        'REQ': zmq.REQ,
        'REP': zmq.REP,
        'ROUTER': zmq.ROUTER,
        'DEALER': zmq.DEALER,
        'PAIR': zmq.PAIR,
    }

    def __init__(self, *,
                 address=None,
                 host=None,
                 port=None,
                 socket_mode,
                 bind,
                 context=None,
                 verbose=True,
                 hwm=None):
        """
        Attributes:
            address
            host
            port
            bind
            socket_mode
            socket_type

        Args:
            address: either specify address or (host and port), but not both
            host: "localhost" is translated to 127.0.0.1
                use "*" to listen to all incoming connections
            port: int, specifies which port to connect / bind to
            socket_mode: zmq.PUSH, zmq.PULL, etc., or their string names
            bind: True -> bind to address, False -> connect to address (see zmq)
            context: Zmq.Context object, if None, creates its own context
            verbose: set to True to print log messages
        """
        # Initialize them first as they are used in __del__
        self.established = False
        self._verbose = verbose

        self._context = None
        self._owns_context = False
        self._set_context(context)

        self.host = None
        self.port = None
        self.address = None
        self._set_address(host, port, address)

        if isinstance(socket_mode, str):
            socket_mode = self.SOCKET_TYPES[socket_mode.upper()]
        self.socket_mode = socket_mode
        self.bind = bind

        self._socket = self._context.socket(self.socket_mode)
        if hwm is not None:
            self._socket.set_hwm(hwm)

    def _set_context(self, context):
        if context is None:
            self._context = zmq.Context()
            self._owns_context = True
        else:
            self._context = context
            self._owns_context = False

    def _set_address(self, host, port, address):
        if address is None:
            # https://stackoverflow.com/questions/6024003/why-doesnt-zeromq-work-on-localhost
            if not ((host is not None) and (port is not None)):
                raise ValueError("either specify address or "
                                 "(host and port), but not both")
            if host == 'localhost':
                host = '127.0.0.1'
            address = "{}:{}".format(host, port)
        try:
            address = str(address)
            if '://' in address:
                self.address = address
            else:  # defaults to TCP
                self.address = 'tcp://' + address
            self.host, port = self.address.split('//')[1].split(':')
            self.port = int(port)
        except Exception as e:
            raise ValueError(
                'Cannot parse address {}. {}'.format(address, str(e)))

    def unwrap(self):
        """
        Get the raw underlying ZMQ socket
        """
        return self._socket

    def establish(self):
        """
        We want to allow subclasses to configure the socket before connecting
        """
        if self.established:
            return
        self.established = True
        if self.bind:
            if self._verbose:
                logger.infofmt('[{}] binding to {}',
                               self.socket_type, self.address)
            self._socket.bind(self.address)
        else:
            if self._verbose:
                logger.infofmt('[{}] connecting to {}',
                               self.socket_type, self.address)
            self._socket.connect(self.address)
        return self

    def __getattr__(self, attrname):
        """
        Delegate any unknown methods to the underlying self.socket
        """
        if attrname in dir(self):
            return object.__getattribute__(self, attrname)
        else:
            return getattr(self._socket, attrname)

    @property
    def socket_type(self):
        """

        Maps name to socket type, i.e. 'PUB' => zmq.PUB

        Returns:
            zmq.<socket_type>
        """
        reverse_map = {value: name for name,
                       value in self.SOCKET_TYPES.items()}
        return reverse_map[self.socket_mode]


class ZmqPusher:
    """
        Pulls data from a ZmqPusher
        Powered by a zmq PUSH socket

        Bind/Connect configurable, by default connects.
    """

    def __init__(self, *,
                 host=None,  # host, port and address can be provided with
                 port=None,  # **kwargs, they are defined here for clarify
                 address=None,
                 bind=False,
                 serializer=None,
                 callback=None,
                 **kwargs):
        """

        Initializes a PUSH socket

        Args:
            host, port, address, bind: See ZmqSocket for details
            serializer: applied to data sent, should return bytes
                (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to ZmqSocket.__init__, along with
                `socket_mode=zmq.PUSH` and `bind = False`
        """
        self.socket = ZmqSocket(host=host,
                                port=port,
                                address=address,
                                socket_mode=zmq.PUSH,
                                bind=bind,
                                **kwargs)
        if callback is not None:
            callback(self.socket)
        self.socket.establish()
        self.serializer = get_serializer(serializer)

    def push(self, data):
        """

        Sends data through the socket

        Args:
            data: Any type supported by the serializer
        """
        data = self.serializer(data)
        self.socket.send(data)


class ZmqPuller:
    """
        Pulls data from a ZmqPusher
        Powered by a zmq PULL socket

        Bind/Connect configurable, by default binds.
    """

    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=True,
                 deserializer=None,
                 callback=None,
                 **kwargs):
        """

        Initializes a PULL socket

        Args:
            host, port, address, bind: See ZmqSocket for details
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        self.socket = ZmqSocket(
            host=host,
            port=port,
            address=address,
            socket_mode=zmq.PULL,
            bind=bind,
            **kwargs,
        )
        if callback is not None:
            callback(self.socket)
        self.socket.establish()
        self.deserializer = get_deserializer(deserializer)

    def pull(self):
        """

        Receives data through the socket

        Returns:
            deserailized data
        """
        data = self.socket.recv()
        return self.deserializer(data)


class ZmqClient:
    """
        Sends request and receives reply from ZmqServer
        Powered by a zmq REQ socket

        By default connects
    """

    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=False,
                 timeout=-1,
                 serializer=None,
                 deserializer=None,
                 callback=None,
                 **kwargs):
        """
        Args:
            timeout: how long do we wait for response, in seconds,
                negative means wait indefinitely
            serializer: applied to data sent, should return bytes
                (default: {None})
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        self.timeout = timeout
        self.address = address
        self.host = host
        self.port = port
        self.bind = bind
        self.serializer = get_serializer(serializer)
        self.deserializer = get_deserializer(deserializer)
        self.callback = callback
        self.kwargs = kwargs
        self._setup()

    def _setup(self):
        self.socket = ZmqSocket(
            address=self.address,
            host=self.host,
            port=self.port,
            socket_mode=zmq.REQ,
            bind=self.bind,
            **self.kwargs
        )
        if self.timeout >= 0:
            self.socket.setsockopt(zmq.LINGER, 0)
            self.poller = zmq.Poller()
            self.poller.register(self.socket.unwrap(), zmq.POLLIN)
        if self.callback is not None:
            self.callback(self.socket)
        self.socket.establish()

    def request(self, msg):
        """
        Send request @msg to self.address

        https://github.com/zeromq/pyzmq/issues/132
        We allow the requester to time out. When timeout happens,
        it will discard the message.

        Args:
            msg: send msg to ZmqServer to request for reply

        Returns:
            reply data from ZmqServer

        Raises:
            ZmqTimeoutError if timed out
        """
        if self.socket.closed:
            self._setup()
        msg = self.serializer(msg)
        self.socket.send(msg)
        if self.timeout >= 0:
            if self.poller.poll(self.timeout * 1000):
                rep = self.socket.recv()
                return self.deserializer(rep)
            else:
                self.socket.close()
                raise ZmqTimeoutError()
        else:
            rep = self.socket.recv()
            return self.deserializer(rep)


class ZmqServer:
    """
    Accepts requests from zmq client and replies back
    Powered by a Zmq REP socket

    By default binds
    """

    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=True,
                 serializer=None,
                 deserializer=None,
                 callback=None,
                 **kwargs):
        """
        Args:
            host, port, address, bind: See ZmqSocket for details
            serializer: applied to data sent, should return bytes
                (default: {None})
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        self.serializer = get_serializer(serializer)
        self.deserializer = get_deserializer(deserializer)
        self.socket = ZmqSocket(
            address=address,
            host=host,
            port=port,
            socket_mode=zmq.REP,
            bind=bind,
            **kwargs
        )
        if callback is not None:
            callback(self.socket)
        self.socket.establish()
        self._thread = None
        self._next_step = 'recv'  # for error checking only
        self._stopped = False

    def recv(self):
        """
        Warnings:
            DO NOT call recv() after you start an event_loop
        """
        if self._next_step != 'recv':
            raise ValueError(
                'recv() and send() must be paired. You can only send() now')
        data = self.socket.recv()
        self._next_step = 'send'
        return self.deserializer(data)

    def send(self, msg):
        """
        Warnings:
            DO NOT call send() after you start an event_loop
        """
        if self._next_step != 'send':
            raise ValueError(
                'send() and recv() must be paired. You can only recv() now')
        msg = self.serializer(msg)
        self.socket.send(msg)
        self._next_step = 'recv'

    def stop(self):
        """
            Stop the event loop. Only safe to call if
            Either there is at least one new message
            Or
            it is called by handler
        """
        self._stopped = True

    def _loop(self, handler):
        while True:
            if self._stopped:
                return
            msg = self.recv()  # request msg from ZmqClient
            reply = handler(msg)
            self.send(reply)

    def start_loop(self, handler, blocking=False):
        """
        Args:
            handler: function that takes an incoming client message
                (deserialized) and returns a reply to client
                (which is serialized and sent)
            blocking: True to block the main program
                False to launch a thread in the background
                and immediately returns

        Returns:
            if non-blocking, returns the created thread that has started
        """
        if blocking:
            self._loop(handler)
        else:
            if self._thread:
                raise RuntimeError('loop is already running')
            self._thread = Thread(target=self._loop, args=[handler])
            self._thread.start()
            return self._thread


class ZmqSender(ZmqClient):
    """
        Uses REQ / REP to send and receive data,
        i.e. a client that only expects b'ack'

        by default connects
    """

    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=False,
                 timeout=-1,
                 serializer=None,
                 callback=None,
                 **kwargs):
        """
        Args:
            host, port, address, bind: See ZmqSocket for details
            serializer: applied to data sent, should return bytes
                (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        super().__init__(
            address=address,
            host=host,
            port=port,
            bind=bind,
            serializer=serializer,
            timeout=timeout,
            callback=callback,
            **kwargs,
            )

    def send(self, msg):
        """
        Sends data to receiver

        Args:
            msg: data to send

        Raises:
            ValueError: If anything other than ack is received, raise
        """
        response = self.request(msg)
        if response != b'ack':
            raise ValueError('ZmqSender did not receive ack on the other end',
                             'Was something wrong?')


class ZmqReceiver(ZmqServer):
    """
        Uses REQ / REP to send and receive data,
        i.e. a client that only expects b'ack'

        by default binds
    """
    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=True,
                 deserializer=None,
                 callback=None,
                 **kwargs):
        """
        Args:
            host, port, address, bind: See ZmqSocket for details
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        super().__init__(
            host=host,
            port=port,
            address=address,
            bind=bind,
            deserializer=deserializer,
            callback=callback,
            **kwargs)

    def recv(self):
        """
            Block then and returns the message sent by ZmqSender

        Returns:
            data: deserialized message
        """
        data = super().recv()
        self.send(b'ack')
        return data

    def _loop(self, handler):
        while True:
            if self._stopped:
                return
            msg = self.recv()  # request msg from ZmqClient
            handler(msg)


class ZmqPub:
    """
        Wrapper for zmq.PUB socket
        For PUB-SUB pattern

        Binds by default
    """

    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=True,
                 serializer=None,
                 callback=None,
                 **kwargs):
        """
        Instantiate the underlying zmq.PUB socket

        Args:
            host, port, address, bind: See ZmqSocket for details
            serializer: applied to data sent, should return bytes
                (default: {None})
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        self.socket = ZmqSocket(
            address=address,
            host=host,
            port=port,
            bind=bind,
            socket_mode=zmq.PUB,
            **kwargs
        )
        self.serializer = get_serializer(serializer)
        if callback is not None:
            callback(self.socket)
        self.socket.establish()

    def pub(self, topic, data):
        """publish data

        Args:
            topic: str / byte
            data: data to send to ZmqSub, serialized when applicable
        """
        topic = str2bytes(topic)
        data = self.serializer(data)
        self.socket.send_multipart([topic, data])


class ZmqSub:
    """
        Wrapper for zmq.SUB socket
        For PUB-SUB pattern

        Connects by default
    """
    def __init__(self, *,
                 host=None,
                 port=None,
                 address=None,
                 bind=False,
                 deserializer=None,
                 callback=None,
                 topic=None,
                 **kwargs):
        """
        Instantiate the underlying zmq.PUB socket

        Args:
            host, port, address, bind: See ZmqSocket for details
            serializer: applied to data sent, should return bytes
                (default: {None})
            deserializer: applied to data received (default: {None})
            callback: if not None, call `callback(socket)` between
                `socket=ZmqSocket()` and `socket.establish()`
                (default: {None})
            topic: see zmq pub sub topic, same as ZmqSub().subscribe(topic)
                (default: {None})
            **kwargs: Passed to `ZmqSocket.__init__`
        """
        self.socket = ZmqSocket(
            address=address,
            host=host,
            port=port,
            socket_mode=zmq.SUB,
            bind=bind,
            **kwargs,
        )
        if topic is not None:
            self.subscribe(topic)
        self.deserializer = get_deserializer(deserializer)
        if callback is not None:
            callback(self.socket)
        self.socket.establish()
        self._thread = None
        self._stopped = False

    def subscribe(self, topic):
        """Add a topic to subscription

        socket.setsocketopt(zmq.SUBSCRIBE, topic)

        Args:
            topic: str / bytes
        """
        topic = str2bytes(topic)
        self.socket.setsockopt(zmq.SUBSCRIBE, topic)

    def unsubscribe(self, topic):
        """Removes a topic from subscription

        socket.setsockopt(zmq.UNSUBSCRIBE, topic)

        Args:
            topic: str / bytes
        """
        topic = str2bytes(topic)
        self.socket.setsockopt(zmq.UNSUBSCRIBE, topic)

    def recv(self):
        """
        receives a published message

        Returns:
            data: whatever ZmqPub sends,
                  deserialized when applicable
        """
        _, data = self.socket.recv_multipart()
        data = self.deserializer(data)
        return data

    def _loop(self, handler):
        while True:
            if self._stopped:
                return
            msg = self.recv()  # request msg from ZmqClient
            handler(msg)

    def start_loop(self, handler, blocking=False):
        """
        Args:
            handler: function that takes an incoming client message
                (deserialized)
            blocking: True to block the main program
                False to launch a thread in the background
                and immediately returns

        Returns:
            if non-blocking, returns the created thread that has started
        """
        if blocking:
            self._loop(handler)
        else:
            if self._thread:
                raise RuntimeError('loop is already running')
            self._thread = Thread(target=self._loop, args=[handler])
            self._thread.start()
            return self._thread



"""
    Defines a proxy that forwards zmq messages
"""
from threading import Thread
from multiprocessing import Process
import zmq
import nanolog as nl


logger = nl.Logger.create_logger(
    __name__,
    stream='stdout',
    time_format='MD HMS',
    show_level=True,
)


class ZmqProxy:
    """
        Creates a proxy that forwards zmq requests.
    """
    def __init__(self, in_add, out_add, context=None, pattern='router-dealer'):
        """
            The initializer collects all necessary information for the proxy
            It does not create any sockets yet
        Args:
            in_add: address that frontend binds to
            out_add: address that backend binds to
            context: Provided context for zmq,
                if None, a private context is created
            pattern: what zmq proxy pattern to use specified by
                <frontend>-<backend> i.e. `router-dealer` or `pull-push`
        """
        self._initialize()
        self.in_add = in_add
        self.out_add = out_add
        self.context = context
        self.pattern = pattern
        if pattern == 'router-dealer':
            self.frontend_protocol = zmq.ROUTER
            self.backend_protocol = zmq.DEALER
        elif pattern == 'pull-push':
            self.frontend_protocol = zmq.PULL
            self.backend_protocol = zmq.PUSH
        else:
            raise ValueError("Unkown zmq proxy patter {}. Please choose "
                             "router-dealer or pull-push".format(pattern))
        if self.context is None:
            self.context = zmq.Context()

    def run(self):
        """
            Run the proxy based on the information provided
        """
        frontend = self.context.socket(self.frontend_protocol)
        frontend.bind(self.in_add)

        # Socket facing services
        backend = self.context.socket(self.backend_protocol)
        backend.bind(self.out_add)

        message = 'Forwarding traffic from {} to {}. Pattern: {}'
        logger.debugfmt(message, self.in_add, self.out_add, self.pattern)
        zmq.proxy(frontend, backend)

        # TODO: termination
        # We never get here…
        frontend.close()
        backend.close()
        self.context.term()

    def _initialize(self):
        """
            Used by subclasses, called in __init__ before everything else
        """
        pass


class ZmqProxyThread(ZmqProxy, Thread):
    """
        Run the proxy in a thread
    """
    def _initialize(self):
        Thread.__init__(self)
        self.daemon = True


class ZmqProxyProcess(ZmqProxy, Process):
    """
        Run the proxy in a process
    """
    def _initialize(self):
        Process.__init__(self)



"""
Serializes numpy and JSON-like objects
"""
import pickle
import base64
import hashlib
import json
import pyarrow
# import marshal


def pa_serialize(obj):
    return pyarrow.serialize(obj).to_buffer()


def pa_deserialize(binary):
    return pyarrow.deserialize(binary)


def bytes2str(bytestring):
    if isinstance(bytestring, str):
        return bytestring
    else:
        return bytestring.decode('UTF-8')


def str2bytes(string):
    if isinstance(string, bytes):
        return string
    else:
        return string.encode('UTF-8')


_SERIALIZERS = {
    'str': str2bytes,
    'pickle': pickle.dumps,
    'json': lambda data: str2bytes(json.dumps(data)),
    'pyarrow': pa_serialize,
}

_DESERIALIZERS = {
    'str': bytes2str,
    'pickle': pickle.loads,
    'json': lambda data: json.loads(bytes2str(data)),
    'pyarrow': pa_deserialize,
}


def _id(x):
    return x


def get_serializer(serializer):
    """
    Args:
        serializer: one of the following
            - pre-set method (str)
            - callable
            - None: no-op
    """
    if serializer is None:
        return _id
    elif isinstance(serializer, str):
        # pre-set serializers
        serializer = serializer.lower()
        return _SERIALIZERS[serializer]
    elif callable(serializer):
        return serializer
    else:
        raise ValueError('serializer must be either a callable or one of {}'
                         .format(list(_SERIALIZERS.keys())))


def get_deserializer(deserializer):
    """
    Args:
        deserializer: one of the following
            - pre-set method (str)
            - callable
            - None: no-op
    """
    if deserializer is None:
        return _id
    elif isinstance(deserializer, str):
        # pre-set serializers
        deserializer = deserializer.lower()
        return _DESERIALIZERS[deserializer]
    elif callable(deserializer):
        return deserializer
    else:
        raise ValueError('deserializer must be either a callable or one of {}'
                         .format(list(_DESERIALIZERS.keys())))


def string_hash(s):
    assert isinstance(s, str)
    return binary_hash(s.encode('utf-8'))


def binary_hash(binary):
    """
    Low collision hash of any binary string
    For designating the 16-char object key in Redis.
    Runs at 200 mu-second per hash on Macbook pro.
    Only contains characters from [a-z][A-Z]+_
    """
    s = hashlib.md5(binary).digest()
    s = base64.b64encode(s)[:16]
    s = s.decode('utf-8')
    # return s.replace('/','_')
    return s


def pyobj_hash(obj, serializer):
    serializer = get_serializer(serializer)
    return binary_hash(serializer(obj))



import os
from setuptools import setup, find_packages


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read().strip()

setup(
    name='caraml',
    version='0.9.post1',
    author='Surreal AI team',
    url='https://github.com/SurrealAI/caraml',
    description='Carefree Accelerated Messaging Library: protocol layer '
                'for distributed machine learning',
    long_description=read('README.rst'),
    keywords=['Distributed', 'Messaging', 'Machine Learning'],
    license='GPLv3',
    packages=[
        package for package in find_packages() if package.startswith("caraml")
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python :: 3"
    ],
    install_requires=[
        "pyarrow",
        "nanolog",
        "zmq",
    ],
    python_requires='>=3.5',
    include_package_data=True,
    zip_safe=False
)



"""

Wrapper for pyarrow based shared memory files

Variables:
    TEMP_FOLDER {str} -- location to put shared memory files
    memory_usage {dict} -- tracks all temporary file, for debug only
"""
import pyarrow as pa
import os
import uuid

# TODO: allow custom tempfolder
TEMP_FOLDER = '/tmp/caraml'
os.makedirs(TEMP_FOLDER, exist_ok=True)
memory_usage = {}


class SharedMemoryObject(object):
    """
    Encapusulates a pyarrow memory backed file
    """
    def __init__(self, filename, debug=False):
        """
            Initializer
        Args:
            filename: location to store the temporary file
            debug: print additional debug information
        """
        self.debug = debug
        if self.debug:
            print('Shared memory with name {} created'.format(filename))
            memory_usage[filename] = True
        self.filename = filename
        self.file = pa.memory_map(filename)
        self.buffer = self.file.read_buffer()
        self.data = pa.deserialize(self.buffer)
        self.deleted = False

    def delete(self):
        """
            Deletes the underlying shared memory file
        """
        if not self.deleted:
            # assert memory_usage[self.filename]
            if self.debug:
                del memory_usage[self.filename]
                print('Shared memory with name {} deleted'.format(self.filename))
                print('Memory entries: {}'.format(len(memory_usage)))
            self.file.close()
            if self.debug:
                print('Deleting file: {}'.format(self.filename))
            os.remove(self.filename)
            self.deleted = True

    def __del__(self):
        self.delete()


def inmem_dump(data, name=None):
    """
        Dump data to a memory mapped file, return filename
    Args:
        data: bytes data
        name (Optional):
    """
    if name is None:
        name = os.path.join(TEMP_FOLDER, str(uuid.uuid4()))
    with pa.MemoryMappedFile.create(name, len(data)) as f:
        f.write(data)
    return name.encode()


def inmem_serialize(data, name=None):
    """
        Serialize data into pyarrow format,
        Save to a memory mapped file, return filename
    Args:
        data: python object to be serialized. 
               At least supports native types, dict, list and numpy.array
               If data is pyarrow.lib.Buffer, saves directly
        name (Optional):
    """
    if name is None:
        name = os.path.join(TEMP_FOLDER, str(uuid.uuid4()))
    buf = pa.serialize(data).to_buffer()
    with pa.MemoryMappedFile.create(name, buf.size) as f:
        f.write(buf)
    return name.encode()


def inmem_deserialize(name_bin):
    """
        Deserialize data sent by inmem_serialize by
        putting it inside a SharedMemoryObject.data
    """
    return SharedMemoryObject(name_bin.decode())
